from django.test import TestCase

from metenox.api.fuzzwork import get_type_ids_prices


class TestFuzzWork(TestCase):

    def test_crash_empty_search_query(self):
        """
        With an empty search query the API returns [] instead of a dict.
        This makes the code crash whe calling .items() on the result
        """

        get_type_ids_prices([])



# flake8: noqa
"""Interactions with moonmining module"""

from collections import defaultdict
from typing import Dict, List

from moonmining.models.moons import Moon as MoonMiningMoon

from eveuniverse.models import EveType

from metenox.app_settings import (
    METENOX_HARVEST_REPROCESS_YIELD,
    METENOX_HOURLY_HARVEST_VOLUME,
)
from metenox.models import Moon as MetenoxMoon

EVE_MOON_MATERIALS_GROUP_ID = 427  # group id of all moon goo


def list_all_moons() -> List[MetenoxMoon]:
    """Returns all known moons"""
    return MetenoxMoon.objects.all()


def get_metenox_hourly_harvest(moon_id: int) -> Dict[EveType, int]:
    """
    Will return how much moon materials a Metenox anchored on the given moon will harvest from the moon ID
    The output dict is {moon_goo_type: moon_goo_amount_per_hour}
    """
    moon = MoonMiningMoon.objects.get(eve_moon_id=moon_id)

    hourly_ore_amounts = {
        moon_product.ore_type: moon_product.amount * METENOX_HOURLY_HARVEST_VOLUME
        for moon_product in moon.products_sorted()
    }

    outputs = defaultdict(int)

    rock_volume = 10
    rock_to_reprocess = 100

    # https://gitlab.com/ErikKalkoken/aa-moonmining/-/blob/master/moonmining/models/extractions.py#L422
    for ore_type, ore_amount in hourly_ore_amounts.items():
        total_rocks = ore_amount / rock_volume
        number_reprocess = total_rocks / rock_to_reprocess

        # type_material is an EveTypeMaterial
        # https://django-eveuniverse.readthedocs.io/en/latest/api.html#eveuniverse.models.EveTypeMaterial
        for type_material in ore_type.materials.select_related(
            "material_eve_type__market_price"
        ):
            if (
                type_material.material_eve_type.eve_group.id
                == EVE_MOON_MATERIALS_GROUP_ID
            ):
                amount_material = int(
                    number_reprocess
                    * METENOX_HARVEST_REPROCESS_YIELD
                    * type_material.quantity
                )
                outputs[type_material.material_eve_type] += amount_material

    return outputs



"""Corporations views"""

from django_datatables_view.base_datatable_view import BaseDatatableView

from django.contrib.auth.decorators import login_required, permission_required
from django.contrib.auth.mixins import LoginRequiredMixin, PermissionRequiredMixin
from django.db import models
from django.http import JsonResponse
from django.shortcuts import get_object_or_404, render

from allianceauth.eveonline.evelinks import dotlan
from app_utils.views import link_html

from metenox.models import HoldingCorporation, Owner
from metenox.views.general import add_common_context

from ._helpers import corporation_notifications_button_html


@login_required
@permission_required("metenox.basic_access")
def list_corporations(request):
    """Will list all corporations and their statistics"""

    return render(request, "metenox/corporations.html", add_common_context())


# pylint: disable = too-many-ancestors
class CorporationListJson(
    PermissionRequiredMixin, LoginRequiredMixin, BaseDatatableView
):
    """Datatable view rendering corporations"""

    model = HoldingCorporation
    permission_required = "metenox.basic_access"
    columns = [
        "id",
        "corporation_name",
        "alliance_name",
        "count_metenoxes",
        "raw_revenue",
        "profit",
        "details",
    ]

    order_columns = [
        "pk",
        "",
        "",
        "",
        "",
        "",
        "",
    ]

    def render_column(self, row, column):
        if column == "id":
            return row.pk

        if column == "alliance_name":
            return self._render_alliance(row)

        if column == "corporation_name":
            return self._render_corporation(row)

        if column == "details":
            return self._render_details(row)

        return super().render_column(row, column)

    def _render_alliance(self, row: HoldingCorporation) -> str:
        alliance = row.corporation.alliance
        if alliance:
            alliance_link = link_html(
                dotlan.alliance_url(alliance.alliance_name), alliance.alliance_name
            )
            return alliance_link
        return ""

    def _render_corporation(self, row: HoldingCorporation) -> str:
        corporation_link = link_html(
            dotlan.corporation_url(row.corporation_name), row.corporation_name
        )
        return corporation_link

    def _render_details(self, row):
        return corporation_notifications_button_html(row)

    def get_initial_queryset(self):
        return self.initial_queryset(self.request)

    @classmethod
    def initial_queryset(cls, request):
        """Basic queryset of the function"""
        holding_corporations_query = HoldingCorporation.objects.select_related(
            "corporation",
            "corporation__alliance",
        ).filter(is_active=True)

        if not request.user.has_perm("metenox.auditor"):
            user_owners = Owner.objects.filter(character_ownership__user=request.user)
            holding_corporations_query = holding_corporations_query.filter(
                owners__in=user_owners
            )

        return holding_corporations_query

    def filter_queryset(self, qs):
        """Use params in the GET to filter"""

        qs = self._apply_search_filter(qs, 0, "corporation__corporation_name")

        qs = self._apply_search_filter(qs, 1, "corporation__alliance__alliance_name")

        return qs

    def _apply_search_filter(self, qs, column_num, field) -> models.QuerySet:
        my_filter = self.request.GET.get(f"columns[{column_num}][search][value]", None)
        if my_filter:
            if self.request.GET.get(f"columns[{column_num}][search][regex]", False):
                kwargs = {f"{field}__iregex": my_filter}
            else:
                kwargs = {f"{field}__istartswith": my_filter}
            return qs.filter(**kwargs)
        return qs


@login_required
@permission_required("metenox.basic_access")
def corporation_fdd_data(request) -> JsonResponse:
    """Provide lists for drop down fields"""
    qs = CorporationListJson.initial_queryset(request)
    columns = request.GET.get("columns")
    result = {}
    if columns:
        for column in columns.split(","):
            options = _calc_options(request, qs, column)
            result[column] = sorted(list(set(options)), key=str.casefold)
    return JsonResponse(result, safe=False)


# pylint: disable = too-many-return-statements, duplicate-code
def _calc_options(request, qs, column):
    if column == "alliance_name":
        values = qs.values_list("corporation__alliance__alliance_name", flat=True)
        values = (value for value in values if value)
        return values

    if column == "corporation_name":
        return qs.values_list("corporation__corporation_name", flat=True)

    return [f"** ERROR: Invalid column name '{column}' **"]


@login_required
@permission_required("metenox.basic_access")
def corporation_notifications(request, corporation_pk: int):
    """Render notification details of a corporation"""

    corporation = get_object_or_404(HoldingCorporation, pk=corporation_pk)
    context = {"corporation": corporation}

    if request.GET.get("new_page"):
        context["title"] = "Corporation notifications"
        context["content_file"] = "metenox/partials/corporation_notifications.html"
        return render(
            request,
            "metenox/modals/generic_modal_page.html",
            add_common_context(context),
        )

    return render(
        request,
        "metenox/modals/corporation_notifications.html",
        add_common_context(context),
    )



"""Models."""

from math import ceil
from typing import Dict, List, Set

from moonmining.models import Moon as MoonminigMoon

from django.contrib.auth.models import User
from django.db import models
from django.db.models import Count, Min, Sum
from django.utils import timezone
from esi.models import Token
from eveuniverse.models import EveGroup, EveMoon, EveType

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveCorporationInfo
from allianceauth.services.hooks import get_extension_logger

from metenox.app_settings import METENOX_MOON_MATERIAL_BAY_CAPACITY

ESI_SCOPES = [
    "esi-universe.read_structures.v1",
    "esi-corporations.read_structures.v1",
    "esi-assets.read_corporation_assets.v1",
]

logger = get_extension_logger(__name__)


class General(models.Model):
    """A meta model for app permissions."""

    class Meta:
        managed = False
        default_permissions = ()
        permissions = (
            ("basic_access", "Can access this app"),
            ("auditor", "Can access metenox information about all corporations"),
        )


class HoldingCorporation(models.Model):
    """Corporation holding metenox moon drills"""

    corporation = models.OneToOneField(
        EveCorporationInfo, on_delete=models.CASCADE, primary_key=True
    )

    is_active = models.BooleanField(default=True)
    last_updated = models.DateTimeField(null=True, default=None)

    @property
    def alliance(self):
        """Returns holding corp's alliance"""
        return self.corporation.alliance

    @property
    def corporation_name(self):
        """Returns the holding corporation's name"""
        return self.corporation.corporation_name

    @property
    def count_metenoxes(self) -> int:
        """Return the number of metenoxes a holding corporation owns"""
        return self.metenoxes.count()

    @property
    def raw_revenue(self) -> float:
        """Returns the raw metenox revenus before fuel prices"""
        return self.metenoxes.aggregate(Sum("moon__value"))["moon__value__sum"]

    @property
    def profit(self) -> float:
        """Returns the metenoxes profit after fuel prices"""
        if self.raw_revenue:
            return self.raw_revenue - self.count_metenoxes * Moon.fuel_price()
        return 0.0

    def __str__(self) -> str:
        return self.corporation_name


class Owner(models.Model):
    """Character in corporation owning metenoxes"""

    corporation = models.ForeignKey(
        HoldingCorporation,
        on_delete=models.CASCADE,
        related_name="owners",
    )

    character_ownership = models.ForeignKey(
        CharacterOwnership,
        on_delete=models.SET_DEFAULT,
        default=None,
        null=True,
        related_name="+",
        help_text="Character used to sync this corporation from ESI",
    )

    is_enabled = models.BooleanField(
        default=True,
        db_index=True,
        help_text="Disabled corporations are excluded from the update process",
    )

    class Meta:
        verbose_name = "Owner"
        verbose_name_plural = "Owners"

    def __str__(self):
        return self.name

    @property
    def name(self) -> str:
        """Return name."""
        alliance_ticker_str = (
            f" [{self.corporation.alliance.alliance_ticker}]"
            if self.corporation.alliance
            else ""
        )
        return f"{self.corporation}{alliance_ticker_str} - {self.character_ownership.character}"

    @property
    def alliance_name(self) -> str:
        """Return alliance name."""
        return (
            self.corporation.alliance.alliance_name if self.corporation.alliance else ""
        )

    def fetch_token(self) -> Token:
        """Return valid token for this mining corp or raise exception on any error."""
        if not self.character_ownership:
            raise RuntimeError("This owner has no character configured.")
        token = (
            Token.objects.filter(
                character_id=self.character_ownership.character.character_id
            )
            .require_scopes(ESI_SCOPES)
            .require_valid()
            .first()
        )
        if not token:
            raise Token.DoesNotExist(f"{self}: No valid token found.")
        return token

    @classmethod
    def get_owners_associated_to_user(cls, user: User):
        """Returns all owners of the user"""
        return cls.objects.filter(character_ownership__user=user)


class Moon(models.Model):
    """Represents a moon and the metenox related values"""

    __HOURLY_MAGMATIC = 55
    __HOURLY_FUEL_BLOCKS = 5

    eve_moon = models.OneToOneField(
        EveMoon, on_delete=models.CASCADE, primary_key=True, related_name="+"
    )

    moonmining_moon = models.OneToOneField(
        MoonminigMoon,
        on_delete=models.CASCADE,
        null=True,  # metenox might be dropped on an unknown moon
        default=None,
        related_name="+",
    )

    value = models.FloatField(default=0)
    value_updated_at = models.DateTimeField(null=True, default=None)

    @property
    def hourly_pull(self) -> Dict[EveType, int]:
        """Returns how much goo is harvested in an hour by a metenox"""
        hourly_products = MetenoxHourlyProducts.objects.filter(moon=self)
        return {product.product: product.amount for product in hourly_products}

    @property
    def name(self) -> str:
        """Returns name of this moon"""
        return self.eve_moon.name.replace("Moon ", "")

    @property
    def rarity_class(self):
        """Returns rarity class of this moon"""
        return self.moonmining_moon.rarity_class

    def update_price(self):
        """Updates the Metenox price attribute to display"""
        hourly_harvest_value = sum(
            EveTypePrice.get_eve_type_price(moon_goo) * moon_goo_amount
            for moon_goo, moon_goo_amount in self.hourly_pull.items()
        )
        self.value = hourly_harvest_value * 24 * 30
        self.value_updated_at = timezone.now()
        self.save()

    @property
    def cycles_before_full(self) -> int:
        """Number of harvest cycles before the moon material bay is at full capacity"""
        bay_capacity = METENOX_MOON_MATERIAL_BAY_CAPACITY
        harvest_volume = sum(
            goo_type.volume * amount for goo_type, amount in self.hourly_pull.items()
        )
        return ceil(bay_capacity / harvest_volume) if harvest_volume != 0 else 0

    @classmethod
    def fuel_price(cls) -> float:
        """Returns the monthly price of running a metenox"""
        hourly_price = (
            cls.__HOURLY_MAGMATIC * EveTypePrice.get_magmatic_gases_price()
            + cls.__HOURLY_FUEL_BLOCKS * EveTypePrice.get_fuel_block_price()
        )
        return hourly_price * 24 * 30

    @classmethod
    def moons_in_need_of_update(cls):
        """
        Returns a queryset of all moons that have products in the moonmining application but none here
        """

        moons = Moon.objects.annotate(
            num_metenox_products=Count("hourly_products", distinct=True),
            num_moonmining_products=Count("moonmining_moon__products", distinct=True),
        )
        return moons.filter(num_metenox_products__lt=1, num_moonmining_products__gt=0)

    def profit(self) -> float:
        """Returns the monthly profit of a meteneox including the fuel price"""
        return self.value - self.fuel_price()

    def __str__(self):
        return self.name


class Metenox(models.Model):
    """
    Represents a metenox anchored on a moon
    """

    structure_id = models.PositiveBigIntegerField(primary_key=True)
    structure_name = models.TextField(max_length=150)

    moon = models.OneToOneField(
        Moon,
        on_delete=models.CASCADE,
        related_name="metenox",
    )
    corporation = models.ForeignKey(
        HoldingCorporation, on_delete=models.CASCADE, related_name="metenoxes"
    )

    fuel_blocks_count = models.IntegerField(default=0)
    magmatic_gas_count = models.IntegerField(default=0)

    def get_stored_moon_materials(self) -> List["MetenoxStoredMoonMaterials"]:
        """Return the list of stored moon materials in that Metenox"""
        return list(self.stored_moon_materials.all())

    def get_stored_moon_materials_value(self) -> float:
        """Return the value of all moon materials stored in the metenox"""
        return sum(
            EveTypePrice.get_eve_type_price(stored_moon_material.product)
            * stored_moon_material.amount
            for stored_moon_material in self.get_stored_moon_materials()
        )

    def get_stored_moon_materials_volume(self) -> float:
        """Return the volume of all moon materials stored in the metenox"""
        return sum(
            stored_moon_material.product.volume
            for stored_moon_material in self.get_stored_moon_materials()
        )

    def __str__(self):
        return self.structure_name

    class Meta:
        verbose_name_plural = "Metenoxes"


class MetenoxHourlyProducts(models.Model):
    """
    Represents how much moon goo a Metenox harvests in an hour
    """

    moon = models.ForeignKey(
        Moon, on_delete=models.CASCADE, related_name="hourly_products"
    )
    product = models.ForeignKey(EveType, on_delete=models.CASCADE, related_name="+")

    amount = models.IntegerField()

    def __str__(self):
        return f"{self.product.name} - {self.amount}"

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["moon", "product"], name="functional_pk_metenoxhourlyproduct"
            )
        ]

    @classmethod
    def all_moon_goo_ids(cls) -> Set[int]:
        """Returns all known moon goo ids in the database"""
        return set(cls.objects.values_list("product", flat=True).order_by("product_id"))


class MetenoxStoredMoonMaterials(models.Model):
    """
    Represents how much moon materials are currently stored inside a Metenox
    """

    metenox = models.ForeignKey(
        Metenox, on_delete=models.CASCADE, related_name="stored_moon_materials"
    )
    product = models.ForeignKey(EveType, on_delete=models.CASCADE, related_name="+")

    amount = models.IntegerField(default=0)

    def __str__(self):
        return f"{self.product.name} - {self.amount} in {self.metenox.structure_name}"

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["metenox", "product"], name="functional_pk_metenoxstoredproduct"
            )
        ]


class EveTypePrice(models.Model):
    """
    Represent an eve type and its last fetched price
    """

    __MOON_GOOS_GROUP_ID = 427
    __FUEL_BLOCK_GROUP_ID = 1136
    __MAGMATIC_TYPE_ID = 81143

    eve_type = models.OneToOneField(EveType, on_delete=models.CASCADE, related_name="+")
    price = models.FloatField(default=0)
    last_update = models.DateTimeField(auto_now=True)

    def __str__(self):
        return f"{self.eve_type.name} - {self.price} ISK"

    def update_price(self, new_price: float):
        """Updates the price of an item"""
        self.price = new_price
        self.save()

    @classmethod
    def get_eve_type_price(cls, eve_type: EveType) -> float:
        """Returns the price of an item"""
        type_price, _ = cls.objects.get_or_create(eve_type=eve_type)
        return type_price.price

    @classmethod
    def get_eve_type_id_price(cls, eve_type_id: int) -> float:
        """Returns the price of an item id"""
        return cls.get_eve_type_price(EveType.objects.get(id=eve_type_id))

    @classmethod
    def _get_type_ids_from_group(cls, group_id: int) -> Set[int]:
        """Fetches type ids from their group and returns it as a set"""
        group = EveGroup.objects.get(id=group_id)
        return set(group.eve_types.filter(published=True).values_list("id", flat=True))

    @classmethod
    def get_fuels_type_ids(cls) -> Set[int]:
        """Fetches the id of all 4 fuel blocks from their group and magmatic gas"""
        return cls._get_type_ids_from_group(cls.__FUEL_BLOCK_GROUP_ID) | {
            cls.__MAGMATIC_TYPE_ID
        }

    @classmethod
    def get_moon_goos_type_ids(cls) -> Set[int]:
        """Fetches the ids of all moon goos from their group"""
        return cls._get_type_ids_from_group(cls.__MOON_GOOS_GROUP_ID)

    @classmethod
    def get_magmatic_gas_type_id(cls) -> int:
        """Return the type of magmatic gases"""
        return cls.__MAGMATIC_TYPE_ID

    @classmethod
    def get_fuel_block_price(cls) -> float:
        """Returns the price of the cheapest fuel block"""
        return (
            cls.objects.filter(eve_type__eve_group=cls.__FUEL_BLOCK_GROUP_ID).aggregate(
                Min("price")
            )["price__min"]
            or 0.0
        )

    @classmethod
    def get_magmatic_gases_price(cls) -> float:
        """Returns the price of a unit of magmatic gases"""
        return cls.get_eve_type_id_price(cls.__MAGMATIC_TYPE_ID)



"""Routes."""

from django.urls import path

from . import views

app_name = "metenox"

urlpatterns = [
    path("", views.list_moons, name="index"),
    path("modal_loader_body", views.modal_loader_body, name="modal_loader_body"),
    path("moon/<int:moon_pk>", views.moon_details, name="moon_details"),
    path("moons_data", views.MoonListJson.as_view(), name="moons_data"),
    path("moons_fdd_data", views.moons_fdd_data, name="moons_fdd_data"),
    path("add_owner", views.add_owner, name="add_owner"),
    path("metenoxes", views.metenoxes, name="metenoxes"),
    path("metenoxes/<int:metenox_pk>", views.metenox_details, name="metenox_details"),
    path("metenoxes_data", views.MetenoxListJson.as_view(), name="metenoxes_data"),
    path("metenoxes_fdd_data", views.metenox_fdd_data, name="metenoxes_fdd_data"),
    path("corporations", views.list_corporations, name="corporations"),
    path(
        "corporations_data",
        views.CorporationListJson.as_view(),
        name="corporations_data",
    ),
    path(
        "corporations_fdd_data",
        views.corporation_fdd_data,
        name="corporations_fdd_data",
    ),
    path(
        "corporation/notification/<int:corporation_pk>",
        views.corporation_notifications,
        name="notifications",
    ),
    path("prices", views.prices, name="prices"),
]



"""Utility views or functions used everywhere"""

from django.shortcuts import render
from django.views.decorators.cache import cache_page

from metenox.models import Moon


def add_common_context(context: dict = None) -> dict:
    """Enhance the templates context with context that should be added to every page"""
    if context is None:
        context = {}

    if basic_title := context.get("page_title"):
        context["page_title"] = f"{basic_title} - Metenox"
    else:
        context["page_title"] = "Metenox"

    context["monthly_fuel_price"] = Moon.fuel_price()

    return context


@cache_page(3600)
def modal_loader_body(request):
    """Draw the loader body. Useful for showing a spinner while loading a modal."""
    return render(request, "metenox/modals/loader_body.html")



# flake8: noqa

from .corporations import (
    CorporationListJson,
    corporation_fdd_data,
    corporation_notifications,
    list_corporations,
)
from .general import modal_loader_body
from .metenoxes import (
    MetenoxListJson,
    add_owner,
    metenox_details,
    metenox_fdd_data,
    metenoxes,
)
from .moons import MoonListJson, list_moons, moon_details, moons_fdd_data
from .prices import prices



"""Admin site."""

from django.contrib import admin

from metenox.models import EveTypePrice, HoldingCorporation, Metenox, Moon, Owner
from metenox.templatetags.metenox import formatisk


@admin.register(Moon)
class MoonAdmin(admin.ModelAdmin):
    list_display = ["name", "moon_value", "value_updated_at"]
    fields = [("eve_moon", "moonmining_moon"), ("value", "value_updated_at")]
    search_fields = ["eve_moon__name"]
    readonly_fields = ["eve_moon", "moonmining_moon"]

    @admin.display(description="value")
    def moon_value(self, moon: Moon):
        return f"{formatisk(moon.value)} ISK"

    # TODO add harvested materials


@admin.register(Metenox)
class MetenoxAdmin(admin.ModelAdmin):
    list_display = [
        "structure_name",
        "corporation",
        "metenox_value",
        "fuel_blocks_count",
        "magmatic_gas_count",
    ]
    fields = [
        ("structure_id", "structure_name"),
        "moon",
        "corporation",
        ("fuel_blocks_count", "magmatic_gas_count"),
    ]
    readonly_fields = ["structure_id", "moon"]

    @admin.display(description="value")
    def metenox_value(self, metenox: Metenox):
        return f"{formatisk(metenox.moon.value)} ISK"


class OwnerCharacterAdminInline(admin.TabularInline):
    model = Owner

    def has_add_permission(self, request, obj):
        return False

    def has_change_permission(self, request, obj=None):
        return False


@admin.register(HoldingCorporation)
class HoldingCorporationAdmin(admin.ModelAdmin):
    list_display = ["corporation", "is_active", "count_metenox"]
    readonly_fields = ["corporation", "last_updated", "count_metenox"]
    inlines = (OwnerCharacterAdminInline,)

    @admin.display(description="Number metenoxes")
    def count_metenox(self, holding: HoldingCorporation) -> int:
        return holding.count_metenoxes


@admin.register(EveTypePrice)
class EveTypePriceAdmin(admin.ModelAdmin):
    list_display = ["eve_type", "type_price"]
    readonly_fields = ["eve_type", "last_update"]

    @admin.display(description="Price")
    def type_price(self, type_price: EveTypePrice):
        return f"{formatisk(type_price.price)} ISK"



# Generated by Django 4.2.15 on 2024-09-05 11:48

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("moonmining", "0007_add_localization"),
        ("eveuniverse", "0010_alter_eveindustryactivityduration_eve_type_and_more"),
        ("eveonline", "0017_alliance_and_corp_names_are_not_unique"),
        ("authentication", "0023_alter_userprofile_language"),
    ]

    operations = [
        migrations.CreateModel(
            name="General",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
            options={
                "permissions": (
                    ("basic_access", "Can access this app"),
                    (
                        "auditor",
                        "Can access metenox information about all corporations",
                    ),
                ),
                "managed": False,
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="HoldingCorporation",
            fields=[
                (
                    "corporation",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        primary_key=True,
                        serialize=False,
                        to="eveonline.evecorporationinfo",
                    ),
                ),
                ("is_active", models.BooleanField(default=True)),
                ("last_updated", models.DateTimeField(default=None, null=True)),
            ],
        ),
        migrations.CreateModel(
            name="Metenox",
            fields=[
                (
                    "structure_id",
                    models.PositiveBigIntegerField(primary_key=True, serialize=False),
                ),
                ("structure_name", models.TextField(max_length=150)),
                ("fuel_blocks_count", models.IntegerField(default=0)),
                ("magmatic_gas_count", models.IntegerField(default=0)),
                (
                    "corporation",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="metenoxes",
                        to="metenox.holdingcorporation",
                    ),
                ),
            ],
            options={
                "verbose_name_plural": "Metenoxes",
            },
        ),
        migrations.CreateModel(
            name="Owner",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "is_enabled",
                    models.BooleanField(
                        db_index=True,
                        default=True,
                        help_text="Disabled corporations are excluded from the update process",
                    ),
                ),
                (
                    "character_ownership",
                    models.ForeignKey(
                        default=None,
                        help_text="Character used to sync this corporation from ESI",
                        null=True,
                        on_delete=django.db.models.deletion.SET_DEFAULT,
                        related_name="+",
                        to="authentication.characterownership",
                    ),
                ),
                (
                    "corporation",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="owners",
                        to="metenox.holdingcorporation",
                    ),
                ),
            ],
            options={
                "verbose_name": "Owner",
                "verbose_name_plural": "Owners",
            },
        ),
        migrations.CreateModel(
            name="Moon",
            fields=[
                (
                    "eve_moon",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        primary_key=True,
                        related_name="+",
                        serialize=False,
                        to="eveuniverse.evemoon",
                    ),
                ),
                ("value", models.FloatField(default=0)),
                ("value_updated_at", models.DateTimeField(default=None, null=True)),
                (
                    "moonmining_moon",
                    models.OneToOneField(
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="moonmining.moon",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="MetenoxStoredMoonMaterials",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("amount", models.IntegerField(default=0)),
                (
                    "metenox",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="stored_moon_materials",
                        to="metenox.metenox",
                    ),
                ),
                (
                    "product",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="eveuniverse.evetype",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="MetenoxHourlyProducts",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("amount", models.IntegerField()),
                (
                    "moon",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="hourly_products",
                        to="metenox.moon",
                    ),
                ),
                (
                    "product",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="eveuniverse.evetype",
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="metenox",
            name="moon",
            field=models.OneToOneField(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="metenox",
                to="metenox.moon",
            ),
        ),
        migrations.CreateModel(
            name="EveTypePrice",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("price", models.FloatField(default=0)),
                ("last_update", models.DateTimeField(auto_now=True)),
                (
                    "eve_type",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="eveuniverse.evetype",
                    ),
                ),
            ],
        ),
        migrations.AddConstraint(
            model_name="metenoxstoredmoonmaterials",
            constraint=models.UniqueConstraint(
                fields=("metenox", "product"), name="functional_pk_metenoxstoredproduct"
            ),
        ),
        migrations.AddConstraint(
            model_name="metenoxhourlyproducts",
            constraint=models.UniqueConstraint(
                fields=("moon", "product"), name="functional_pk_metenoxhourlyproduct"
            ),
        ),
    ]



import unittest

from metenox.models import EveTypePrice

from .testdata.load_eveuniverse import load_eveuniverse


class TestPrices(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()

    def test_fuel_block_price_not_return_none(self):
        """In case fuel block prices aren't in the database return 0 instead of None"""

        fuel_block_price = EveTypePrice.get_fuel_block_price()

        self.assertIsNotNone(fuel_block_price)
        self.assertEqual(fuel_block_price, 0.0)

    def test_get_price_on_empty_db(self):
        """If no prices are present in the database 0 should be returned instead of anything else"""

        magmatic_price = EveTypePrice.get_magmatic_gases_price()

        self.assertIsNotNone(magmatic_price)
        self.assertEqual(magmatic_price, 0.0)



from collections import defaultdict

from moonmining.models import Moon, MoonProduct

from django.test import TestCase
from eveuniverse.models import EveType

from metenox.models import Moon as MetenoxMoon
from metenox.moons import get_metenox_hourly_harvest

from ..tasks import create_moon_from_moonmining, update_moons_from_moonmining
from .testdata.load_eveuniverse import load_eveuniverse

MOON_ID = 40178441


class TestMoons(TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()

    def test_wh_moon_hourly_pull(self):
        """Hourly pull of a wormhole moon (only r4)"""
        # TODO
        self.assertTrue(True)

    def test_kspace_moon_hourly_pull(self):
        """Hourly pull of an R16 moon"""
        moon_scan: dict[int, float] = {
            45501: 0.303093851,
            45497: 0.088907719,
            45499: 0.307283372,
            45490: 0.300715059,
        }

        moon, _ = Moon.objects.get_or_create(eve_moon_id=MOON_ID)
        moon_products = [
            MoonProduct(moon=moon, ore_type_id=ore_type_id, amount=ore_type_amount)
            for ore_type_id, ore_type_amount in moon_scan.items()
        ]

        moon.update_products(moon_products)

        harvest = get_metenox_hourly_harvest(moon.eve_moon_id)

        expected = defaultdict(int)
        expected.update(
            {
                EveType.objects.get(id=16634): 234,
                EveType.objects.get(id=16641): 145,
                EveType.objects.get(id=16635): 36,
                EveType.objects.get(id=16633): 36,
                EveType.objects.get(id=16644): 147,
                EveType.objects.get(id=16637): 42,
            }
        )

        self.assertDictEqual(harvest, expected)

    def test_double_r64_hourly_pull(self):
        """
        Test on a moon with double r64
        Way more moon materials present there
        """

        moon_scan = {
            45512: 0.210578531,
            45497: 0.272598475,
            45499: 0.296382815,
            45510: 0.220440164,
        }

        moon, _ = Moon.objects.get_or_create(eve_moon_id=MOON_ID)
        moon_products = [
            MoonProduct(moon=moon, ore_type_id=ore_type_id, amount=ore_type_amount)
            for ore_type_id, ore_type_amount in moon_scan.items()
        ]

        moon.update_products(moon_products)

        harvest = get_metenox_hourly_harvest(moon.eve_moon_id)

        expected = defaultdict(int)
        expected.update(
            {
                EveType.objects.get(id=16634): 52,
                EveType.objects.get(id=16640): 52,
                EveType.objects.get(id=16650): 58,
                EveType.objects.get(id=16635): 35,
                EveType.objects.get(id=16633): 50,
                EveType.objects.get(id=16644): 167,
                EveType.objects.get(id=16652): 55,
                EveType.objects.get(id=16639): 50,
                EveType.objects.get(id=16637): 130,
                EveType.objects.get(id=16642): 26,
            }
        )

        self.assertDictEqual(harvest, expected)

    def test_integrity_error_on_updating_existing_moon(self):
        """
        Integrity error issue when running an update on an already existing moon
        """

        moon_scan = {
            45490: 0.2809781134,
            45491: 0.2083230466,
            45493: 0.5106988549,
        }

        moonmining_moon, _ = Moon.objects.get_or_create(eve_moon_id=MOON_ID)
        moon_products = [
            MoonProduct(
                moon=moonmining_moon, ore_type_id=ore_type_id, amount=ore_type_amount
            )
            for ore_type_id, ore_type_amount in moon_scan.items()
        ]

        moonmining_moon.update_products(moon_products)

        create_moon_from_moonmining(MOON_ID)
        create_moon_from_moonmining(MOON_ID)

    def test_create_moon_mats_on_existing_moon(self):
        """
        When a moon is created without moon materials `create_or_update_moon_from_moonmining` needs to create them
        """

        moon_scan = {
            45490: 0.2809781134,
            45491: 0.2083230466,
            45493: 0.5106988549,
        }

        moonmining_moon, _ = Moon.objects.get_or_create(eve_moon_id=MOON_ID)
        moon_products = [
            MoonProduct(
                moon=moonmining_moon, ore_type_id=ore_type_id, amount=ore_type_amount
            )
            for ore_type_id, ore_type_amount in moon_scan.items()
        ]

        moonmining_moon.update_products(moon_products)
        metenox_moon = MetenoxMoon.objects.create(
            eve_moon_id=MOON_ID,
            moonmining_moon=moonmining_moon,
        )

        update_moons_from_moonmining()

        pull = metenox_moon.hourly_pull

        self.assertEqual(len(pull), 3)

    def test_find_moons_that_need_update(self):
        """
        Will check if it's correctly finding moons with products in the moonmining application but none in the metenox application
        """

        moon_scan = {
            45490: 0.2809781134,
            45491: 0.2083230466,
            45493: 0.5106988549,
        }

        moonmining_moon, _ = Moon.objects.get_or_create(eve_moon_id=MOON_ID)

        metenox_moon = MetenoxMoon.objects.create(
            eve_moon_id=MOON_ID,
            moonmining_moon=moonmining_moon,
        )

        moon_products = [
            MoonProduct(
                moon=moonmining_moon, ore_type_id=ore_type_id, amount=ore_type_amount
            )
            for ore_type_id, ore_type_amount in moon_scan.items()
        ]
        moonmining_moon.update_products(moon_products)

        moon_to_update = MetenoxMoon.moons_in_need_of_update()

        self.assertEqual(list(moon_to_update), [metenox_moon])

    def test_no_crash_cycles_before_full_empty_moon(self):
        """
        Checks that a moon without a survey in moonmining won't crash when cycles_before_full is called
        """

        moonmining_moon, _ = Moon.objects.get_or_create(eve_moon_id=MOON_ID)

        metenox_moon = MetenoxMoon.objects.create(
            eve_moon_id=MOON_ID,
            moonmining_moon=moonmining_moon,
        )

        metenox_moon.cycles_before_full



"""
Modules containing all ESI interactions
"""

from collections import defaultdict
from typing import Dict, List, Optional, Set

from esi.clients import EsiClientProvider
from esi.models import Token

from metenox.models import HoldingCorporation

from . import __version__

METENOX_TYPE_ID = 81826


esi = EsiClientProvider(app_info_text=f"aa-metenox v{__version__}")


def get_metenox_from_esi(
    holding_corporation: HoldingCorporation,
) -> Optional[List[Dict]]:
    """Returns all metenoxes associated with a given Owner"""

    structures = get_structures_from_esi(holding_corporation)

    return (
        [
            structure
            for structure in structures
            if structure["type_id"] == METENOX_TYPE_ID
        ]
        if structures
        else None
    )


def get_structure_info_from_esi(
    holding_corporation: HoldingCorporation, structure_id: int
) -> Dict:
    """Returns the location information of a structure"""

    for owner in holding_corporation.owners.all():

        structure_info = esi.client.Universe.get_universe_structures_structure_id(
            structure_id=structure_id,
            token=owner.fetch_token().valid_access_token(),
        ).result()

        return structure_info


def get_structures_from_esi(
    holding_corporation: HoldingCorporation,
) -> Optional[List[Dict]]:
    """Returns all structures associated with a given owner"""

    for owner in holding_corporation.owners.all():
        try:
            return esi.client.Corporation.get_corporations_corporation_id_structures(
                corporation_id=owner.corporation.corporation.corporation_id,
                token=owner.fetch_token().valid_access_token(),
            ).results()
        except Token.DoesNotExist:  # tries the next owner
            continue

    return None  # No owner worked


def get_corporation_assets(holding_corporation: HoldingCorporation):
    """Returns all the assets of a corporation"""

    for owner in holding_corporation.owners.all():
        try:
            return esi.client.Assets.get_corporations_corporation_id_assets(
                corporation_id=holding_corporation.corporation.corporation_id,
                token=owner.fetch_token().valid_access_token(),
            ).results()
        except Token.DoesNotExist:  # tries the next owner
            continue

    return None


def get_corporation_metenox_assets(
    holding_corporation: HoldingCorporation, metenoxes_set_ids: Set[int]
) -> Optional[Dict[int, List[Dict]]]:
    """
    Return the assets in the corporation's Metenoxes' MoonMaterialBay and FuelBay.
    Need to receive the set of the corporation's metenoxes ids.
    The data is formatted as a dict with the key being the metenox structure id and a list with the info
    """

    holding_assets = get_corporation_assets(holding_corporation)
    assets_dic = defaultdict(list)
    for asset in holding_assets:
        if asset["location_id"] in metenoxes_set_ids and asset["location_flag"] in [
            "MoonMaterialBay",
            "StructureFuel",
        ]:
            assets_dic[asset["location_id"]].append(asset)

    return assets_dic



from django.core.management import BaseCommand, call_command

from allianceauth.services.hooks import get_extension_logger

logger = get_extension_logger(__name__)


class Command(BaseCommand):
    help = "Preload static data from the ESI"

    def handle(self, *args, **options):
        call_command(
            "eveuniverse_load_types",
            "metenox",
            "--group_id",
            427,  # moon goo
            "--group_id",
            1136,  # fuel block
            "--type_id",
            81143,  # magmatic
        )

        """
        self.stdout.write("Loading data from the ESI. It might take a while")
        goo_types_count = EveType.objects.filter(eve_group_id=25).count()
        self.stdout.write(f"Already {goo_types_count} in the database.")
        self.stdout.write()
        user_input = input("Do you want to proceed? (y/N)?")

        if user_input.lower() == "y":
            self.stdout.write("Starting to fetch moon goos")
            update_or_create_eve_object.delay(
                model_name="EveType",
                id=25,
                enabled_sections=[
                    EveType.Section.DOGMAS,
                    EveType.Section.TYPE_MATERIALS,
                ],
            )

            self.stdout.write("Done")

        else:
            self.stdout.write("Aborted")
        """
